// C Source File
// Created 7/17/2010; 12:00:38 PM

// Tokenizer for Solar89
// Converts a text file into an array of hexadecimel tokens

//General includes
#include <alloc.h>
#include <dialogs.h>
#include <kbd.h>
#include <mem.h>
#include <statline.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <vat.h>

#include "tokeninfo.h"

//NUM_TOKENS value here... this does not have to be changed since it is set when tokenizePrgm is called
int NUM_TOKENS = 0;

//Create a "null" ? token (0xAF)
tokenInfo createNullToken()
{
	unsigned long rawhex;
	tokenInfo nullToken;
	
	//Set the properties of the null token
	nullToken.twoByteToken = 0;
	nullToken.chars = (size_t)1;
	rawhex = strtol("0xAF", NULL, 0);
	nullToken.hex[0] = (unsigned char)rawhex;

	return nullToken;
}

//Create a hard return token (0x3F)
tokenInfo createRetToken()
{
	unsigned long rawhex;
	tokenInfo nullToken;
	
	//Set the properties of the null token
	nullToken.twoByteToken = 0;
	nullToken.chars = (size_t)1;
	rawhex = strtol("0x3F", NULL, 0);
	nullToken.hex[0] = (unsigned char)rawhex;

	return nullToken;
}

//Check if a string passed to it is a token, returns the place of that token in the array
int isToken(char *string, tokenInfo *tokens)
{
	int i;
	int tokenNum = -1;
	
	tokenInfo token;
	
	//Loop over string tokens
	for (i = 0; i < NUM_TOKENS; ++i)
	{
		token = tokens[i];
		if (strcmp(token.name, string) == 0)
			tokenNum = i;
	}
	
	return tokenNum;
}

//Given a string, this should find the longest token and return its length and hex
tokenInfo findToken(char *string, tokenInfo *tokens)
{
	int i;
	int tokenNum;
	int length = strlen(string);

	char *truncated = NULL;
	char *errormsg = NULL;

	tokenInfo token = createNullToken();

	//Allocate memory
	if ((errormsg = (char *)calloc(75, sizeof(char))) == NULL) 
	{
		DlgMessage("Alloc Failure", "Unable to allocate space for error message", BT_OK, BT_NONE);
		return token;
	}
	if ((truncated = (char *)calloc(10, sizeof(char))) == NULL) 
	{
		DlgMessage("Alloc Failure", "Unable to allocate space for truncated string", BT_OK, BT_NONE);
		free(errormsg);
		return token;
	}

	//If the entire thing is a token, deal with it now
	tokenNum = isToken(string, tokens);
	if (tokenNum != -1)
	{
		token = tokens[tokenNum];
		free(errormsg);
		free(truncated);
		return token;
	}
	else if (strlen(string) == 1)
	{
		sprintf(errormsg, "The character \"%s\" is not defined as a token in the selected token file", string);
		DlgMessage("Error", errormsg, BT_OK, BT_NONE);
		free(errormsg);
		free(truncated);
		return token;
	}

	//Loop over the characters of the string
	for (i = 0; i < length; ++i)
	{
		//Realloc memory to the truncated string and set it
		truncated = memset(truncated, '\0', strlen(truncated));
		truncated = (char *)realloc(truncated, length * sizeof(char));
		truncated = strncat(truncated, string, (length-1)-i);
		
		//Check if it's a token- if it is, return, if not and if the length is one, return the null token
		tokenNum = isToken(truncated, tokens);
		if (tokenNum != -1)
		{
			token = tokens[tokenNum];
			break;
		}
		else if (strlen(truncated) == 1)
		{
			sprintf(errormsg, "The character \"%s\" is not defined as a token in the selected token file", truncated);
			DlgMessage("Error", errormsg, BT_OK, BT_NONE);
			break;
		}
	}
	
	//Return the token and free memory
	free(truncated);
	free(errormsg);
	return token;
}

//Takes a string and a buffer to store generated hex (thanks to calcdude84se!), returns the number of swaps
int tokenizeString(char *string, unsigned char *buffer, tokenInfo *tokens) 
{
	int numswaps = 0;

	tokenInfo currentToken;
	tokenInfo retToken;

	//Then, if the line begins with whitespace, get rid of it
	while (string[0] == ' ')
		string += 1;

	//Manipulate the string pointer in order to find all tokens it contains
	while (*string != '\0') 
	{
		currentToken = findToken(string, tokens);
		
		//Handle the token if it is two bytes first (due to endian issues)
		if (currentToken.twoByteToken == 1)
		{
			*(buffer++) = currentToken.hex[1];
			*(buffer++) = currentToken.hex[0];
			numswaps += 2;
		}
		//If it's not, perform the much simpler operation
		else
		{
			*(buffer++) = currentToken.hex[0];
			numswaps += 1;
		}

		//Increment pointer
		string += currentToken.chars;
	}

	//Add the hard return token now, at the end of the line
	retToken = createRetToken();
	*(buffer++) = retToken.hex[0];
	numswaps += 1;

	return numswaps;
}

//Convert a string (from a text file) consisting of tokens in string format, into an array of hexadecimel tokens
void tokenizePrgm(char *input, unsigned char *output, tokenInfo *tokens, int numtokens)
{
	int span = 0;
	int length = strlen(input);
	int numswaps = 0;
	int i;
	
	char newchar;
	char endline = '\r';
	char *line = NULL;
	char *working = NULL;
	unsigned char *tokline = NULL;
	
	NUM_TOKENS = numtokens;
	
	//Allocate memory to the strings
	if ((working = (char *)calloc(5, sizeof(char))) == NULL) 
	{
		DlgMessage("Alloc Failure", "Unable to allocate space for working string", BT_OK, BT_NONE);
		return;
	}
	if ((tokline = (unsigned char *)calloc(10, sizeof(unsigned char))) == NULL)
	{
		DlgMessage("Alloc Failure", "Unable to allocate space for token line string", BT_OK, BT_NONE);
		free(working);
		return;		
	}
	if ((line = (char *)calloc(10, sizeof(char))) == NULL) 
	{
		DlgMessage("Alloc Failure", "Unable to allocate space for line string", BT_OK, BT_NONE);
		free(tokline);
		free(working);
		return;
	}
	
	//Loop until the length of the input string is zero
	while (length > 0)
	{
		//Get the length of the line up to trailing character, memset and reallocate memory
		sprintf(working, "%c", endline);
		span = strcspn(input, working);
		line = memset(line, '\0', strlen(line));
		line = (char *)realloc(line, (span + 2) * sizeof(char));
		tokline = memset(tokline, '\0', strlen(tokline));
		tokline = (char *)realloc(tokline, (span + 2) * sizeof(unsigned char));
		
		//Load the line into a string, and then remove it
		line = strncat(line, input, span);
		input = strpbrk(input, working);
		if (span == 0)
			input += 2;

		//Then clear out the \r at the front of the string and clear the working string out
		newchar = input[2];
		sprintf(working, "%c", newchar);
		input = strpbrk(input, working);
		length = strlen(input);

		//Pass the line, the output array, and the token info array to tokenizeString, keep track of place in output array
		numswaps = tokenizeString(line, tokline, tokens);
		for (i = 0; i < numswaps; ++i)
			*(output++) = tokline[i];
	}
	
	//Free memory
	free(line);
	free(working);
	free(tokline);
}