#!/usr/bin/env python

"""Parses non-group *.89_ files

program file for the TI-82, TI-83, TI-85, or TI-86, instead of the default,

This Python module has functions to parse a non-group *.89* file (a file 
from the TI-89 graphing calculator, or any other TI calculator using a 
Motorola 68000 CPU (collectively called 68k calcs).

The code is designed to sort through the irrelevant information (such as
bytes always set to one value, or null bytes) and return the 'important'
information.
"""

def getName(text):
	"""Returns the name of a *.89_ variable file"""
	name = text[64:72]
	return name

def getFolder(text):
	"""Returns the default folder of a *.89_ variable file"""
	folder = text[10:18]
	return folder

def getType(text):
	"""Returns the default type of a *.89_ variable file"""
	type = text[72:73]
	return type
	
def getData(text):
	"""Returns the actual data, in string form, of a *.89_ variable file"""
	end = len(text) - 4
	data = text[88:end]		#This number may need changing!
	return data

def open89var(filename):
	"""Opens an 89 variable and returns a tuple of its name, default folder, 
	data type, and data (as string)"""
	
	#Try to open it, if impossible raise an IO Error
	try:
		file = open(filename, 'rb')
	except:
		raise IOError

	#Read in the data and then close the file
	output = file.read()
	file.close()
	
	#Retrieve information
	name = getName(output)
	folder = getFolder(output)
	type = getType(output)
	data = getData(output)
	
	#Return a tuple of the information
	return name, folder, type, data